import boto3
from ezaws.utils.timing import (
    epoch_minutes_ago,
    epoch_days_ago,
    epoch_seconds_ago,
    epoch_hours_ago,
)
from botocore.exceptions import ClientError
from ezaws.models.cloudwatch import (
    CreateLogGroupResponse,
    DeleteLogGroupResponse,
    LogResponse,
    GetLogStreamRespone,
    Event,
    TailLogResponse,
    LogEvent,
    Events,
)
from typing import Optional, Dict, NewType, Any, Sequence, Union, List
from pprint import pprint
import time
from pydantic import BaseModel, Field

from typing_extensions import TypedDict
from ezaws.exceptions import CloudWatchException

StreamName = NewType("StreamName", str)
SequenceToken = NewType("SequenceToken", str)
StreamDict = TypedDict(
    "StreamDict",
    {
        "StreamName": StreamName,
        "SequenceToken": Union[SequenceToken, None],
    },
)


class Log(BaseModel):
    """Interface to Cloudwatch Logs.

    The Log represents a single log group in a region. It can
     perform most common operations, such as:
    - create the log group
    - create streams
    - retrieve logs


    name: name of the log group
    region: the region the log is in
    default_stream_name: name of the stream to log to when the stream name is not
     passed in as an argument to logging functions
    streams: the streams that have been created for the log group
    """

    name: str
    region: str
    default_stream_name: StreamName = "general"  # type: ignore
    streams: Dict[StreamName, Union[SequenceToken, None]] = Field(default_factory=dict)

    def __init__(self, **data: Any):
        """Think '__post_init__'"""
        super().__init__(**data)
        self.set_log_streams()

    def create_log_group(
        self, tags: Optional[Dict[str, str]] = None
    ) -> CreateLogGroupResponse:
        client = boto3.client("logs", region_name=self.region)
        args: dict[str, Any] = {"logGroupName": self.name}
        if tags:
            args["tags"] = tags
        response = client.create_log_group(**args)
        return CreateLogGroupResponse(**response)

    def delete_log_group(self) -> DeleteLogGroupResponse:
        client = boto3.client("logs", region_name=self.region)
        response = client.delete_log_group(logGroupName=self.name)
        return DeleteLogGroupResponse(**response)

    def create_stream(self, stream_name: StreamName) -> None:
        """Create a stream for this Cloudwatch logs instance."""
        client = boto3.client("logs", region_name=self.region)
        client.create_log_stream(logGroupName=self.name, logStreamName=stream_name)
        self.streams[stream_name] = None

    def log(
        self,
        *,
        message: str,
        stream_name: Optional[StreamName] = None,
    ) -> LogResponse:
        """Log a message to a stream.

        Use the default stream is the stream name is not specified."""

        events = [
            LogEvent(message=message, timestamp=int(round(time.time() * 1000))),
        ]
        resp = self.log_events(events=events, stream_name=stream_name)

        return resp

    def log_messages(
        self,
        *,
        messages: List[str],
        stream_name: Optional[StreamName] = None,
    ) -> LogResponse:
        """Log multiple message to a stream.

        Use the default stream is the stream name is not specified."""

        events = []
        for message in messages:
            events.append(
                {
                    "timestamp": int(round(time.time() * 1000)),
                    "message": message,
                }
            )
        resp = self.log_events(events=events, stream_name=stream_name)
        return resp

    def log_events(
        self,
        *,
        events: Sequence[Union[LogEvent, Dict]],
        stream_name: Optional[StreamName] = None,
    ) -> LogResponse:
        """Log multiple message to a stream.

        Use the default stream is the stream name is not specified.

        The events can either be a of the type Event, or they can be dicts,
         like so:
        {
            "timestamp": int(round(time.time() * 1000)),
            "message": "msg1",
        }
        """
        if stream_name is None:
            stream_name = self._get_stream_name()
        client = boto3.client("logs", region_name=self.region)
        # pprint(events)
        events_to_log = list(
            event.dict() for event in events if isinstance(event, LogEvent)
        )
        events_to_log.extend(
            list((event for event in events if isinstance(event, dict)))
        )
        # pprint(events)
        log_event = {
            "logGroupName": self.name,
            "logStreamName": stream_name,
            "logEvents": events_to_log,
        }

        if self.streams.get(stream_name) is not None:
            log_event["sequenceToken"] = self.streams.get(stream_name)  # type: ignore

        response = client.put_log_events(**log_event)

        ret = LogResponse(**response)
        token: SequenceToken = ret.nextSequenceToken  # type: ignore
        self.streams[stream_name] = token
        return ret

    def get_log_streams(
        self,
    ) -> GetLogStreamRespone:
        """Retrieve and store information from all streams for this log group"""
        client = boto3.client("logs", region_name=self.region)
        response = client.describe_log_streams(
            logGroupName=self.name,
        )
        print(response)
        return GetLogStreamRespone(**response)

    def set_log_streams(self) -> None:
        """Retrieve all log streams for this log group and set the next token."""
        try:
            resp = self.get_log_streams()
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                # this means that we do not have any streams yet.
                return None
            raise e

        for stream in resp:
            # print(stream.uploadSequenceToken, stream.logStreamName)
            if isinstance(stream.uploadSequenceToken, int):
                self.streams[stream.logStreamName] = str(stream.uploadSequenceToken)  # type: ignore
            else:
                self.streams[stream.logStreamName] = None
        return None

    def _get_stream_name(self) -> StreamName:
        """Helper to return the default stream name
        and raise an exception in case it does not exist.
        """
        stream_name = self.default_stream_name
        if stream_name not in self.streams.keys():
            raise CloudWatchException(
                f"default_stream_name {stream_name} does not exist. Existing streams:\n{self.streams}"
            )
        return stream_name

    def tail_log(
        self, n: int = 100, stream_name: Optional[str] = None
    ) -> TailLogResponse:
        """Look at latest n events in target stream.

        Checks the default stream if no stream_name is given."""
        if stream_name is None:
            stream_name = self._get_stream_name()
        client = boto3.client("logs", region_name=self.region)

        response = client.get_log_events(
            logGroupName=self.name,
            logStreamName=stream_name,
            # startTime=int(datetime.datetime(2021, 8, 19, 0, 0).strftime('%s'))*1000,
            # endTime=int(datetime.datetime(2021, 8, 20, 0, 0).strftime('%s'))*1000,
            limit=n,
            startFromHead=False,
        )
        return TailLogResponse(**response)

    def get_log_events(
        self,
        stream_name: Optional[str] = None,
        limit: int = 10,
        startTime: Optional[int] = None,
        endTime: Optional[int] = None,
        start_from_head: bool = True,
    ) -> Events:
        """Retrieve the entire log and returns it.

        Checks the default stream if no stream_name is given.

        Returns all the events for the log.

        The last item in the list is the newest event.

        Notes to self:

          - when startFromHead is set to True, it reads the oldest logs first.

          - when startFromHead is set to False, it reads the newest logs first.

          - using 'nextForwardToken' is moving from old to new logs.

          - using 'nextBackwardToken' is moving from new to old logs.
        """
        if stream_name is None:
            stream_name = self._get_stream_name()
        client = boto3.client("logs", region_name=self.region)
        events = Events()
        get_log_events_kwargs = {
            "logGroupName": self.name,
            "logStreamName": stream_name,
            "limit": limit,
            "startFromHead": start_from_head,
        }
        if startTime:
            get_log_events_kwargs["startTime"] = startTime

        if endTime:
            get_log_events_kwargs["endTime"] = endTime

        response = client.get_log_events(**get_log_events_kwargs)
        response = TailLogResponse(**response)
        for event in response.events:
            events.events.append(event)

        # print(response.nextBackwardToken)
        # print(response.nextForwardToken)

        next_token = response.nextForwardToken
        while True:
            get_log_events_kwargs["nextToken"] = next_token
            response = client.get_log_events(**get_log_events_kwargs)
            response = TailLogResponse(**response)
            for event in response.events:
                events.events.append(event)
                print(event)
            # The log is depleted when AWS starts returning
            # the same token over and over.
            if next_token == response.nextForwardToken:
                break
            else:
                next_token = response.nextForwardToken

        return events

    def get_log_events_last_seconds(self, seconds: int) -> Events:
        epoch_seconds = epoch_seconds_ago(seconds)
        events = self.get_log_events(
            limit=5000, startTime=epoch_seconds, start_from_head=True
        )
        return events

    def get_log_events_last_minutes(self, minutes: int) -> Events:
        epoch_minutes = epoch_minutes_ago(minutes)
        events = self.get_log_events(
            limit=5000, startTime=epoch_minutes, start_from_head=True
        )
        return events

    def get_log_events_last_hours(self, hours: int) -> Events:
        epoch_hours = epoch_hours_ago(hours)
        events = self.get_log_events(
            limit=5000, startTime=epoch_hours, start_from_head=True
        )
        return events

    def get_log_events_last_days(self, days: int) -> Events:
        epoch_days = epoch_days_ago(days)
        events = self.get_log_events(
            limit=5000, startTime=epoch_days, start_from_head=True
        )
        return events



import boto3
from boto3.s3.transfer import TransferConfig, S3Transfer
from ezaws.models.s3 import (
    ListBucketsResponse,
    CreateBucketResponse,
    DeleteBucketResponse,
    VersioningResponse,
    ObjectMetadata,
    PutObjectResponse,
    DeleteObjectResponse,
    TCBuilder,
)
from typing import Any, Union, Literal, Optional, Dict
from dataclasses import dataclass


KB = 1024
MB = KB * KB


class S3:
    def list_buckets(self) -> ListBucketsResponse:
        """List all buckets using S3 client."""
        s3 = boto3.client(
            "s3",
        )

        response = s3.list_buckets()
        return ListBucketsResponse(**response)


@dataclass
class S3Bucket:
    region: str
    name: str

    def create_bucket(self) -> CreateBucketResponse:
        """Create the bucket for this class."""
        s3_client = boto3.client("s3")
        resp_bucket_creation = s3_client.create_bucket(
            Bucket=self.name,
            CreateBucketConfiguration={"LocationConstraint": self.region},
        )

        return CreateBucketResponse(**resp_bucket_creation)

    def delete_bucket(self) -> DeleteBucketResponse:
        """Delete this bucket"""
        s3_resource = boto3.resource("s3")
        delete_response = s3_resource.Bucket(self.name).delete()
        return DeleteBucketResponse(**delete_response)

    def get_versioning(self) -> VersioningResponse:
        """Return versioning information for this bucket."""
        s3_client = boto3.client("s3")
        versioning_response = s3_client.get_bucket_versioning(Bucket=self.name)

        return VersioningResponse(**versioning_response)

    def get_object_metadata(self, s3_file_name: str) -> ObjectMetadata:
        """
        Display metadata for target file stored in S3
        """
        s3 = boto3.client("s3")
        response = s3.head_object(Bucket=self.name, Key=s3_file_name)
        return ObjectMetadata(**response)

    def get_file_size(
        self, s3_file_name: str, block_size: Literal["bytes", "MB"] = "MB"
    ) -> int:
        """
        List a file from S3
        """
        resp = self.get_object_metadata(s3_file_name=s3_file_name)
        size = resp.ContentLength

        if block_size == "bytes":
            return int(size)
        elif block_size == "MB":
            return int(size / MB)

    def put_object(
        self, file_name: str, s3_key_name: Optional[str] = None
    ) -> PutObjectResponse:
        """Put an object in S3"""
        s3_key_name = s3_key_name if s3_key_name is not None else file_name
        s3 = boto3.client("s3")

        resp = s3.put_object(Body=file_name, Bucket=self.name, Key=s3_key_name)
        return PutObjectResponse(**resp)

    def upload_file(self, file_name: str, s3_key_name: Optional[str] = None) -> Any:
        """Upload a file to S3 using upload_file

        TODO: examine all of the upload_file possible returns.

        add optional argument to satisfy 'ExtraArgs={'Metadata': {'mykey': 'myvalue'}'
         from https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.S3Transfer.ALLOWED_UPLOAD_ARGS
        """
        s3_key_name = s3_key_name if s3_key_name is not None else file_name
        s3 = boto3.client("s3")

        resp = s3.upload_file(file_name, self.name, s3_key_name)

        return resp

    def controlled_upload(
        self,
        file_name: str,
        s3_key_name: Optional[str] = None,
        tc: Optional[Union[TCBuilder, Dict]] = None,
    ) -> None:

        """
        Upload a file to S3 with more fine-grained control.

        tc is a Union of either of the following:
          - None: in which case the TransferConfig() defaults are used
          - TCBuilder: a model for dev ergonmics that you can use to
           instantiate an instance of TransferConfig()
          - Dict: a dict that is used to instantiate an instance of TransferConfig()

        The TransferConfig description:
         https://boto3.amazonaws.com/v1/documentation/api/latest/_modules/boto3/s3/transfer.html#TransferConfig

        TransferConfig example (default values):

            TransferConfig(
                multipart_threshold=8 * MB,
                max_concurrency=10,
                multipart_chunksize=8 * MB,
                num_download_attempts=5,
                max_io_queue=100,
                io_chunksize=256 * KB,
                use_threads=True,
                max_bandwidth=None,
            )
        """
        s3_key_name = s3_key_name if s3_key_name is not None else file_name

        session = boto3.Session()
        if tc is None:
            tc_config = TransferConfig()  # use default
        elif isinstance(tc, TCBuilder):
            tc_config = TransferConfig(
                **{k: v for (k, v) in tc.dict().items() if v is not None}
            )
        elif isinstance(tc, dict):
            tc_config = TransferConfig(**tc)

        s3_client = session.client("s3")
        transfer_object = S3Transfer(client=s3_client, config=tc_config)
        transfer_object.upload_file(file_name, self.name, s3_key_name)

    def delete_object(
        self, s3_key_name: str, version_id: Optional[str] = None
    ) -> DeleteObjectResponse:
        """Deletes an object from the bucket."""
        s3 = boto3.client("s3")
        del_args = {"Bucket": self.name, "Key": s3_key_name}
        if version_id:
            del_args.update({"VersionID": version_id})

        response = s3.delete_object(**del_args)
        return DeleteObjectResponse(**response)

    def empty(self) -> None:
        raise NotImplementedError

    def set_encryption(self) -> None:
        raise NotImplementedError

    def set_versioning(self) -> None:
        raise NotImplementedError



import boto3


from ezaws.models.sqs import (
    GetQueueResponse,
    CreateQueueResponse,
    DeleteQueueResponse,
    ListQueueResponse,
    PurgeQueueResponse,
    SendMessageResponse,
    ReadMessageResponse,
    DeleteMessageResponse,
)


def create_queue(queue_name: str, region: str) -> CreateQueueResponse:
    """ "Create an SQS queue."""
    client = boto3.client("sqs", region_name=region)
    response = client.create_queue(
        QueueName=queue_name,
        Attributes={
            "DelaySeconds": "5",  # default 0, max 900 (15 min)
            "VisibilityTimeout": "29",  # default 30, min 0
        },
    )
    return CreateQueueResponse(**response)


def get_queue_url(queue_name: str) -> GetQueueResponse:
    sqs_client = boto3.client("sqs", region_name="eu-central-1")
    response = sqs_client.get_queue_url(
        QueueName=queue_name,
    )
    return GetQueueResponse(**response)


def delete_queue(queue_name: str, region: str) -> DeleteQueueResponse:
    sqs_client = boto3.client("sqs", region_name=region)
    response = sqs_client.delete_queue(QueueUrl=queue_name)
    return DeleteQueueResponse(**response)


def purge_queue(queueu_url: str, region_name: str) -> PurgeQueueResponse:
    sqs_client = boto3.client("sqs", region_name=region_name)
    response = sqs_client.purge_queue(QueueUrl=queueu_url)
    return PurgeQueueResponse(**response)


def send_message(
    queueu_url: str, region_name: str, message: str
) -> SendMessageResponse:
    sqs_client = boto3.client("sqs", region_name=region_name)

    response = sqs_client.send_message(QueueUrl=queueu_url, MessageBody=message)
    return SendMessageResponse(**response)


def read_message(
    queueu_url: str, region_name: str, nr_of_messages: int = 1
) -> ReadMessageResponse:
    sqs_client = boto3.client("sqs", region_name=region_name)
    response = sqs_client.receive_message(
        QueueUrl=queueu_url,
        MaxNumberOfMessages=nr_of_messages,  # max 10, default 1
        WaitTimeSeconds=0,
    )
    return ReadMessageResponse(**response)


def delete_message(
    queueu_url: str, region_name: str, receipt_handle: str
) -> DeleteMessageResponse:
    sqs_client = boto3.client("sqs", region_name=region_name)
    response = sqs_client.delete_message(
        QueueUrl=queueu_url,
        ReceiptHandle=receipt_handle,
    )

    return DeleteMessageResponse(**response)


def list_queues(region: str, max_results: int = 100) -> ListQueueResponse:
    """Returns a list of URLs inside a ListQueueResponse,
    listing all the queues in a region."""
    sqs_client = boto3.client("sqs", region_name=region)

    response = sqs_client.list_queues(MaxResults=max_results)
    return ListQueueResponse(**response)



from ezaws.models.common import ResponseMetadata
from pydantic import BaseModel
from typing import Iterator, List, Optional, Any
from ezaws.utils.timing import (
    convert_to_local,
    epoch_to_date_time,
    datetime_to_epoch_in_ms,
)

import datetime


class CreateLogGroupResponse(BaseModel):
    ResponseMetadata: ResponseMetadata


class DeleteLogGroupResponse(BaseModel):
    ResponseMetadata: ResponseMetadata


class LogResponse(BaseModel):
    ResponseMetadata: ResponseMetadata
    nextSequenceToken: str


class LogStreams(BaseModel):
    """Individual stream inside a log group."""

    arn: str
    creationTime: int
    firstEventTimestamp: Optional[int]
    lastEventTimestamp: Optional[int]
    lastIngestionTime: Optional[int]
    logStreamName: str
    storedBytes: int
    uploadSequenceToken: Optional[int]


class GetLogStreamRespone(BaseModel):
    ResponseMetadata: ResponseMetadata
    logStreams: List[LogStreams]

    def __iter__(self) -> Any:
        return iter(self.logStreams)


class LogEvent(BaseModel):
    timestamp: int  # number of milliseconds after Jan 1, 1970
    message: str  # message that should appear in CloudWatch logs


class Event(BaseModel):

    timestamp: int  # number of milliseconds after Jan 1, 1970
    message: str  # message that should appear in CloudWatch logs
    ingestionTime: Optional[int]

    @property
    def date_time_local(self) -> datetime.datetime:
        """Converts the epoch that is returned by Cloudwatch to
        a datetime.datetime object that represents the local time."""
        date_time = epoch_to_date_time(self.timestamp / 1000.0)
        local_date_time = convert_to_local(date_time)
        return local_date_time

    @property
    def epoch_local(self) -> int:
        """Converts the epoch that is returned by Cloudwatch to
        an epoch value that is representative of the local time."""
        date_time = epoch_to_date_time(self.timestamp / 1000.0)
        local_date_time = convert_to_local(date_time)
        local_epoch = datetime_to_epoch_in_ms(local_date_time)
        return local_epoch


class TailLogResponse(BaseModel):
    ResponseMetadata: ResponseMetadata
    events: List[Event]
    nextBackwardToken: str
    nextForwardToken: str

    def __iter__(self) -> Any:
        return iter(self.events)


class Events(BaseModel):
    events: List[Event] = []



"""ez imports"""
from ezaws.models.sqs import Message
from ezaws.sqs.messenger import Messenger
from ezaws.models.regions import Region
from ezaws.ssm.parameter_store import ParameterStore
from ezaws.s3.s3 import S3, S3Bucket
from ezaws.models.s3 import TCBuilder
from ezaws.cloudwatch.logs import Log
from ezaws.models.cloudwatch import LogEvent
from ezaws.models.rds import DBInstanceType, DBEngine
from ezaws.rds.rds import RDS
from ezaws.lambdas.lambdas import Lambda
from ezaws.dynamodb.dynamodb import DynamoDB
from ezaws.models.dynamodb import Table

__all__ = [
    "Messenger",
    "Message",
    "Region",
    "ParameterStore",
    "S3",
    "S3Bucket",
    "TCBuilder",
    "Log",
    "LogEvent",
    "DBInstanceType",
    "DBEngine",
    "RDS",
    "Lambda",
    "DynamoDB",
    "Table",
]



from typing import Dict, Optional, Union
from pydantic import BaseModel


class ResponseMetadata(BaseModel):
    RequestId: str
    HostId: Optional[str]
    HTTPStatusCode: int
    HTTPHeaders: Dict[str, Union[str, bytes]]
    RetryAttempts: Optional[int]



import boto3
from dataclasses import dataclass
from typing import Any
from ezaws.models.sqs import (
    DeleteQueueResponse,
    CreateQueueResponse,
    ListQueueResponse,
    GetQueueResponse,
    DeleteMessageResponse,
    SendMessageResponse,
    ReadMessageResponse,
    PurgeQueueResponse,
    QueueAttributes,
)


# TODO: add read_all() method


@dataclass
class Messenger:

    queueu_url: str
    region: str

    @classmethod
    def from_queue_name(
        cls,
        queue_name: str,
        region: str,
        delay_seconds: str = "0",
        visibility_timeout: str = "30",
    ) -> Any:
        """Returns a Messenger for a region from a 'queue_name'.

        If the queue already exists, that queue is returned. If the
         queue does not exist already, a new queue is created.

        Example:
            >>> x = Messenger.from_queue_name("some_queue", region="eu-central-1")
        """
        list_q_response = cls._list_queues(region, max_results=100)
        for url in list_q_response:
            if url.endswith(queue_name):
                sqs_client = boto3.client("sqs", region_name=region)
                response = sqs_client.get_queue_url(
                    QueueName=queue_name,
                )
                response = GetQueueResponse(**response)
                return Messenger(queueu_url=response.QueueUrl, region=region)
        client = boto3.client("sqs", region_name=region)
        response = client.create_queue(
            QueueName=queue_name,
            Attributes={
                "DelaySeconds": delay_seconds,  # default 0, max 900 (15 min)
                "VisibilityTimeout": visibility_timeout,  # default 30, min 0
            },
        )
        response = CreateQueueResponse(**response)
        return Messenger(queueu_url=response.QueueUrl, region=region)

    def _get_queue_url(
        self,
        queue_name: str,
        region_name: str,
        delay_seconds: str = "0",
        visibility_timeout: str = "30",
    ) -> str:
        """Returns a queue with target name for the region.

        Creates the queue if it does not already exist."""

        list_queue_response = self._list_queues(region_name)
        for q_url in list_queue_response:
            if queue_name in q_url:
                # queue exists, so we return the queueu url:
                return queue_name
        client = boto3.client("sqs", region_name=region_name)
        response = client.create_queue(
            QueueName=queue_name,
            Attributes={
                "DelaySeconds": delay_seconds,  # default 0, max 900 (15 min)
                "VisibilityTimeout": visibility_timeout,  # default 30, min 0
            },
        )
        create_queue_response = CreateQueueResponse(**response)
        return create_queue_response.QueueUrl

    def send_message(self, message: str) -> SendMessageResponse:
        """Send a message to the queue."""
        sqs_client = boto3.client("sqs", region_name=self.region)

        response = sqs_client.send_message(
            QueueUrl=self.queueu_url, MessageBody=message
        )
        return SendMessageResponse(**response)

    def read_messages(self, nr_of_messages: int = 1) -> ReadMessageResponse:
        """Read a number of messages from the queue."""
        sqs_client = boto3.client("sqs", region_name=self.region)
        response = sqs_client.receive_message(
            QueueUrl=self.queueu_url,
            MaxNumberOfMessages=nr_of_messages,  # max 10, default 1
            WaitTimeSeconds=0,
        )
        return ReadMessageResponse(**response)

    def delete_message(self, receipt_handle: str) -> DeleteMessageResponse:
        """Delete target message from the queueu"""
        sqs_client = boto3.client("sqs", region_name=self.region)
        response = sqs_client.delete_message(
            QueueUrl=self.queueu_url,
            ReceiptHandle=receipt_handle,
        )
        return DeleteMessageResponse(**response)

    def get_queue_count(self) -> int:
        """Returns the 'ApproximateNumberOfMessages' from the queueu."""

        return int(self.get_queue_attributes().ApproximateNumberOfMessages)

    def get_queue_attributes(self) -> QueueAttributes:
        """Returns the queueu attributes."""
        sqs_client = boto3.client("sqs", region_name=self.region)
        response = sqs_client.get_queue_attributes(
            QueueUrl=self.queueu_url,
            AttributeNames=["All"],
        )
        ret = QueueAttributes(**response["Attributes"])
        return ret

    def purge_queue(self) -> PurgeQueueResponse:
        """Purge the queue."""
        sqs_client = boto3.client("sqs", region_name=self.region)
        response = sqs_client.purge_queue(QueueUrl=self.queueu_url)

        return PurgeQueueResponse(**response)

    def delete_queue(self) -> DeleteQueueResponse:
        """Deletes the queue"""
        sqs_client = boto3.client("sqs", region_name=self.region)
        response = sqs_client.delete_queue(QueueUrl=self.queueu_url)
        return DeleteQueueResponse(**response)

    @staticmethod
    def _list_queues(region: str, max_results: int = 100) -> ListQueueResponse:
        """Returns a list of URLs inside a ListQueueResponse,
        listing all the queues in a region."""

        sqs_client = boto3.client("sqs", region_name=region)

        response = sqs_client.list_queues(MaxResults=max_results)
        return ListQueueResponse(**response)



from dataclasses import dataclass


@dataclass(frozen=True)
class Region:
    """All the AWS regions"""

    us_east_2: str = "us-east-2"  # US East (Ohio)
    us_east_1: str = "us-east-1"  # US East (N. Virginia)
    us_west_1: str = "us-west-1"  # US West (N. California)
    us_west_2: str = "us-west-2"  # US West (Oregon)
    af_south_1: str = "af-south-1"  # Africa (Cape Town)
    ap_east_1: str = "ap-east-1"  # Asia Pacific (Hong Kong)
    ap_southeast_3: str = "ap-southeast-3"  # Asia Pacific (Jakarta)
    ap_south_1: str = "ap-south-1"  # Asia Pacific (Mumbai)
    ap_northeast_3: str = "ap-northeast-3"  # Asia Pacific (Osaka)
    ap_northeast_2: str = "ap-northeast-2"  # Asia Pacific (Seoul)
    ap_southeast_1: str = "ap-southeast-1"  # Asia Pacific (Singapore)
    ap_southeast_2: str = "ap-southeast-2"  # Asia Pacific (Sydney)
    ap_north_east_1: str = "ap-northeast-1"  # Asia Pacific (Tokyo)
    ca_central_1: str = "ca-central-1"  # Canada (Central)
    cn_north_1: str = "cn-north-1"  # China (Beijing)
    cn_north_west: str = "cn-northwest-1"  # China (Ningxia)
    eu_central_1: str = "eu-central-1"  # Europe (Frankfurt)
    eu_west_1: str = "eu-west-1"  # Europe (Ireland)
    eu_west_2: str = "eu-west-2"  # Europe (London)
    eu_south_1: str = "eu-south-1"  # Europe (Milan)
    eu_west_3: str = "eu-west-3"  # Europe (Paris)
    eu_north_1: str = "eu-north-1"  # Europe (Stockholm)
    me_south_1: str = "me-south-1"  # Middle East (Bahrain)
    sa_east_1: str = "sa-east-1"  # South America (São Paulo)



from pydantic import BaseModel
import datetime
from ezaws.models.common import ResponseMetadata
from typing import List, Optional, Any

"""
response = client.put_parameter(
    Name='string',
    Description='string',
    Value='string',
    Type='String'|'StringList'|'SecureString',
    KeyId='string',
    Overwrite=True|False,
    AllowedPattern='string',
    Tags=[
        {
            'Key': 'string',
            'Value': 'string'
        },
    ],
    Tier='Standard'|'Advanced'|'Intelligent-Tiering',
    Policies='string',
    DataType='string'
)
"""


class CreateParameter(BaseModel):
    Name: str
    Description: Optional[str]
    Value: Optional[str]
    Type: str
    KeyId: Optional[str]
    Overwrite: bool = False
    Policies: Optional[List]
    Tier: Optional[str]
    AllowedPattern: Optional[str]

    def generate_parameter_args(self) -> dict:
        """Generate the Parameter arguments for use by removing all
        keys that have the value 'None'."""
        cur_d = vars(self)
        args = {}
        for key, value in cur_d.items():
            if value:
                args[key] = value
        return args


class Parameter(BaseModel):
    ARN: Optional[str]
    DataType: str
    LastModifiedDate: datetime.datetime
    LastModifiedUser: Optional[str]
    Name: str
    Type: str
    Value: Optional[str]
    Version: int
    Policies: Optional[List]
    Tier: Optional[str]
    AllowedPattern: Optional[str]
    Description: Optional[str]
    KeyId: Optional[str]


class GetParameterResponse(BaseModel):
    Parameter: Parameter
    ResponseMetadata: ResponseMetadata


class RegionParameters(BaseModel):
    Parameters: List[Parameter]
    ResponseMetadata: ResponseMetadata

    def __iter__(self) -> Any:
        return iter(self.Parameters)


class DeteleParameterResponse(BaseModel):
    ResponseMetadata: ResponseMetadata


class CreateParameterResponse(BaseModel):
    Version: int
    Tier: Optional[str]
    ResponseMetadata: ResponseMetadata



from ezaws.models.common import ResponseMetadata
from pydantic import BaseModel
import datetime
from typing import List, Dict, Optional, Literal, Any, Iterator


class BucketBrief(BaseModel):
    Name: str
    CreationDate: datetime.datetime


class ListBucketsResponse(BaseModel):
    ResponseMetadata: ResponseMetadata
    Buckets: List[BucketBrief]
    Owner: Dict[str, str]

    class Config:
        arbitrary_types_allowed = True

    def __iter__(self) -> Any:
        return iter(self.Buckets)


class CreateBucketResponse(BaseModel):
    ResponseMetadata: ResponseMetadata
    Location: str


class DeleteBucketResponse(BaseModel):
    ResponseMetadata: ResponseMetadata


class VersioningResponse(BaseModel):
    ResponseMetadata: ResponseMetadata
    Status: Optional[Literal["Enabled", "Suspended"]]
    MFADelete: Optional[Literal["Enabled", "Disabled"]]
    versioning: bool = False

    def __post_init__(self) -> None:
        if self.Status == "Enabled":
            self.versioning = True


class ObjectMetadata(BaseModel):
    ResponseMetadata: ResponseMetadata
    AcceptRanges: str
    ContentLength: int
    ContentType: str
    ETag: str
    LastModified: datetime.datetime
    Metadata: Dict[Any, Any]
    SSEKMSKeyId: Optional[str]
    ServerSideEncryption: Optional[str]
    VersionId: Optional[str]


class PutObjectResponse(BaseModel):
    ResponseMetadata: ResponseMetadata
    ETag: Optional[str]
    Expiration: Optional[str]
    ChecksumCRC32: Optional[str]
    ChecksumCRC32C: Optional[str]
    ChecksumSHA1: Optional[str]
    ChecksumSHA256: Optional[str]
    ServerSideEncryption: Optional[Literal["AES256", "aws:kms"]]
    VersionId: Optional[str]
    SSECustomerAlgorithm: Optional[str]
    SSECustomerKeyMD5: Optional[str]
    SSEKMSKeyId: Optional[str]
    SSEKMSEncryptionContext: Optional[str]
    BucketKeyEnabled: Optional[bool]
    RequestCharged: Optional[str]


class DeleteObjectResponse(BaseModel):
    ResponseMetadata: ResponseMetadata
    DeleteMarker: Optional[bool]
    VersionId: Optional[str]
    RequestCharged: Optional[str]


class TCBuilder(BaseModel):
    """
    Serves as arg ot the TransferConfig:
      https://boto3.amazonaws.com/v1/documentation/api/latest/_modules/boto3/s3/transfer.html#TransferConfig

    TransferConfig example (default values):

            TransferConfig(
                multipart_threshold=8 * MB,
                max_concurrency=10,
                multipart_chunksize=8 * MB,
                num_download_attempts=5,
                max_io_queue=100,
                io_chunksize=256 * KB,
                use_threads=True,
                max_bandwidth=None,
            )
    """

    multipart_threshold: Optional[int]
    max_concurrency: Optional[int]
    multipart_chunksize: Optional[int]
    use_threads: Optional[bool]
    num_download_attempts: Optional[int]
    max_io_queue: Optional[int]
    io_chunksize: Optional[int]
    max_bandwidth: Optional[int]



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Iterator


@dataclass
class ResponseMetadata:
    RequestId: str
    HTTPStatusCode: int
    HTTPHeaders: Dict[str, str]
    RetryAttempts: Optional[int]


@dataclass
class QueueAttributes:
    ApproximateNumberOfMessages: int
    ApproximateNumberOfMessagesDelayed: int
    ApproximateNumberOfMessagesNotVisible: int
    CreatedTimestamp: int
    DelaySeconds: int
    LastModifiedTimestamp: int
    MaximumMessageSize: int
    MessageRetentionPeriod: int
    QueueArn: str
    ReceiveMessageWaitTimeSeconds: int
    SqsManagedSseEnabled: bool
    VisibilityTimeout: int


@dataclass
class Message:
    MessageId: str
    ReceiptHandle: str
    MD5OfBody: str
    Body: str  # the actual message


@dataclass
class ReadMessageResponse:
    ResponseMetadata: ResponseMetadata
    Messages: List[Message] = field(default_factory=list)

    def __iter__(self) -> Iterator[Message]:
        return iter(self.Messages)


@dataclass
class SendMessageResponse:
    MD5OfMessageBody: str
    MessageId: str
    ResponseMetadata: ResponseMetadata


@dataclass
class DeleteResponse:
    ResponseMetadata: ResponseMetadata
    RequestId: str
    HTTPStatusCode: int
    RetryAttempts: int
    HTTPHeaders: Dict[str, str] = field(default_factory=dict)


@dataclass
class GetQueueResponse:
    QueueUrl: str
    ResponseMetadata: ResponseMetadata


@dataclass
class CreateQueueResponse:
    QueueUrl: str
    ResponseMetadata: ResponseMetadata


@dataclass
class DeleteQueueResponse:
    ResponseMetadata: ResponseMetadata


@dataclass
class ListQueueResponse:
    # ["https://eu-central-1.queue.amazonaws.com/717687450252/example-queue"]
    QueueUrls: List[str]
    ResponseMetadata: ResponseMetadata

    def __iter__(self) -> Iterator[str]:
        return iter(self.QueueUrls)


@dataclass
class PurgeQueueResponse:
    ResponseMetadata: ResponseMetadata


@dataclass
class GetQueueAttributesResponse:
    ResponseMetadata: ResponseMetadata
    Attributes: QueueAttributes


@dataclass
class DeleteMessageResponse:
    ResponseMetadata: ResponseMetadata



from ezaws.models.common import ResponseMetadata
from pydantic import BaseModel
from typing import List, Any, Dict, Optional
from enum import Enum


class DBInstanceType(Enum):
    db_t2_micro = "db.t2.micro"


class DBEngine(Enum):
    MYSQL = "mysql"
    POSTGRES = "postgress"
    MARIA_DB = "mariadb"
    AURORA = "aurora"


class DBInstanceNotFoundResponse(BaseModel):
    error_message: str = "DBInstanceNotFound"


class Endpoint(BaseModel):
    Address: str
    HostedZoneId: str
    Port: int


class DBInstance(BaseModel):
    DBInstanceIdentifier: str
    DBInstanceClass: str
    Engine: str
    DBInstanceStatus: str
    MasterUsername: str
    AllocatedStorage: int
    PreferredBackupWindow: str
    BackupRetentionPeriod: int
    DBSecurityGroups: List[Any]
    VpcSecurityGroups: Dict[Any, Any]
    DBParameterGroups: List[Any]
    DBSubnetGroup: Dict[Any, Any]
    PreferredMaintenanceWindow: str
    PendingModifiedValues: Dict[Any, Any]
    MultiAZ: bool
    EngineVersion: str
    AutoMinorVersionUpgrade: bool
    ReadReplicaDBInstanceIdentifiers: List[Any]
    LicenseModel: str
    OptionGroupMemberships: List[Any]
    PubliclyAccessible: bool
    StorageType: str
    DbInstancePort: int
    StorageEncrypted: bool
    DbiResourceId: str
    CACertificateIdentifier: str
    DomainMemberships: List[Any]
    Endpoint: Optional[Endpoint]
    CopyTagsToSnapshot: bool
    MonitoringInterval: int
    DBInstanceArn: str
    IAMDatabaseAuthenticationEnabled: bool
    PerformanceInsightsEnabled: bool
    DeletionProtection: bool
    AssociatedRoles: List[Any]
    TagList: List[Any]
    CustomerOwnedIpEnabled: bool
    ActivityStreamStatus: Optional[str]
    BackupTarget: str


class DescribeDBResponse(BaseModel):
    ResponseMetadata: ResponseMetadata
    DBInstances: List[DBInstance]


class CreateDBResponse(BaseModel):
    ResponseMetadata: ResponseMetadata
    DBInstance: DBInstance


class DeleteDBResponse(BaseModel):
    DBInstance: DBInstance
    ResponseMetadata: ResponseMetadata


class StartRDSResponse(BaseModel):
    DBInstance: DBInstance
    ResponseMetadata: ResponseMetadata


class StopRDSResponse(BaseModel):
    DBInstance: DBInstance
    ResponseMetadata: ResponseMetadata



from ezaws.models.common import ResponseMetadata
from pydantic import BaseModel
from enum import Enum

from typing import Union, Optional, List, Literal, Dict, Any, Iterator


class AttributeType(Enum):
    S = "S"  # string
    N = "N"  # number
    B = "B"  # binary


class Attribute(BaseModel):
    attribute_name: str
    attribute_type: AttributeType

    class Config:
        use_enum_values = True


# class Attributes(BaseModel):
#    attributes: List[Attribute]
#
#    def iter(self) -> Iterator[Attribute]:
#        return iter(self.attributes)


class Key(Enum):
    HASH = "HASH"  # partition key
    RANGE = "RANGE"  # sort key


class KeySchema(BaseModel):
    key_name: str
    key_type: Key

    class Config:
        use_enum_values = True


# class Keys(BaseModel):
#    keys: List[KeySchema]
#
#    def iter(self) -> Iterator[KeySchema]:
#        return iter(self.keys)


class Table(BaseModel):
    table_name: str
    attributes: List[Attribute] = []
    keys: List[KeySchema] = []
    rcu: int  # ReadCapacityUnits: 1 is free
    wcu: int  # WriteCapacityUnits: 1 is free

    def add_attribute(
        self,
        attribute_name: str,
        attribute_type: Literal["S", "N", "B"],
        key_type: Literal["HASH", "RANGE"],
    ) -> None:
        """Add an attribute to the Table and define the attribute type and key.

        Attribute types:
            "S" for string
            "N" for number
            "B" for binary

        Attribute key types:
            "HASH" for partition key
            "RANGE" for sort key
        """
        attribute = Attribute(
            attribute_name=attribute_name, attribute_type=attribute_type
        )
        key_schema = KeySchema(key_name=attribute_name, key_type=key_type)
        self.attributes.append(attribute)
        self.keys.append(key_schema)


class ProvisionedThroughput(BaseModel):
    NumberOfDecreasesToday: int
    ReadCapacityUnits: int
    WriteCapacityUnits: int


class TableDescription(BaseModel):
    ItemCount: int
    ProvisionedThroughput: ProvisionedThroughput
    TableArn: str
    TableId: str
    TableName: str
    TableSizeBytes: int
    TableStatus: str


class DeleteResponse(BaseModel):
    ResponseMetadata: ResponseMetadata
    TableDescription: TableDescription


class GetItemResponse(BaseModel):
    Item: Optional[dict]
    ResponseMetadata: ResponseMetadata


class ScanResult(BaseModel):
    Count: int
    Items: list[dict]


class QueryResult(BaseModel):
    Count: int
    Items: list
    ResponseMetadata: ResponseMetadata


class PartiQLResult(BaseModel):
    Items: list
    ResponseMetadata: ResponseMetadata


if __name__ == "__main__":

    table = Table(table_name="Humans", rcu=1, wcu=1)
    table.add_attribute(attribute_name="name", attribute_type="S", key_type="RANGE")
    table.add_attribute(attribute_name="id", attribute_type="N", key_type="HASH")

    print(table.dict())



from ezaws.models.common import ResponseMetadata
from pydantic import BaseModel
from typing import List, Dict, Any, Optional, Iterator, Union


class Function(BaseModel):
    Architectures: List[str]
    CodeSha256: str
    CodeSize: int
    Description: str
    Environment: Optional[Dict[str, Any]] = None
    FunctionArn: str
    FunctionName: str
    Handler: str
    LastModified: str
    MemorySize: int
    PackageType: str
    RevisionId: str
    Role: str
    Runtime: str
    Timeout: int
    TracingConfig: Dict[str, str]
    Version: str


class ListFunctionsResponse(BaseModel):
    Functions: List[Function]
    ResponseMetadata: ResponseMetadata

    def __iter__(self) -> Any:
        return iter(self.Functions)


class RunFunctionResponse(BaseModel):
    ExecutedVersion: str
    LogResult: Optional[Union[str, bytes]] = None
    ResponseMetadata: ResponseMetadata
    StatusCode: int
    Payload: Any

    def payload_to_str(self) -> str:
        ret_str = ""
        for some_bytes in self.Payload:
            ret_str += some_bytes.decode("utf-8")
        return ret_str



import boto3
import botocore
from ezaws import Region
from dataclasses import dataclass
from ezaws.models.lambdas import ListFunctionsResponse, RunFunctionResponse
from typing import Union, Optional, Literal, Any
import base64


@dataclass
class Lambda:
    """
    Interface to Lambda service.
    """

    region: Optional[Region] = None
    client: botocore.client = boto3.client("lambda")

    def __post_init__(self) -> None:
        if self.region:
            self.client = boto3.client("lambda", region_name=self.region)

    def list_functions(self) -> ListFunctionsResponse:
        """List all functions for a region."""
        response = self.client.list_functions()
        return ListFunctionsResponse(**response)

    def run_function(
        self,
        function_name: str,
        payload: Optional[Union[str, bytes]] = None,
        invocation: Optional[Literal["RequestResponse", "Event", "DryRun"]] = None,
        log_type: Optional[Literal["Tail"]] = "Tail",
    ) -> RunFunctionResponse:

        args: dict[Any, Any] = {
            "FunctionName": function_name,
        }
        if invocation:
            args["InvocationType"] = invocation
        if log_type:
            args["LogType"] = log_type
        if payload:
            args["Payload "] = payload

        response = self.client.invoke(**args)
        func_response = RunFunctionResponse(**response)
        if func_response.LogResult:
            func_response.LogResult = base64.b64decode(response.LogResult)

        if func_response.ResponseMetadata.HTTPHeaders.get("x-amz-log-result"):
            func_response.ResponseMetadata.HTTPHeaders[
                "x-amz-log-result"
            ] = base64.b64decode(
                func_response.ResponseMetadata.HTTPHeaders["x-amz-log-result"]
            )

        return func_response



import time
import datetime
from typing import Optional
from dateutil import tz


def current_ms_after_epoch() -> int:
    """Returns the current epoch in ms."""
    ms = int(time.time() * 1000)
    return ms


def datetime_to_epoch_in_ms(date: datetime.datetime) -> int:
    """Converts given datetime.datetime instance to epoch in ms."""
    seconds = date.timestamp()
    return int(seconds * 1000)


def epoch_seconds_ago(n: int) -> int:
    """Returns the epoch in ms as it was exactly n-seconds ago."""
    date_time = datetime.datetime.now() - datetime.timedelta(seconds=n)
    return datetime_to_epoch_in_ms(date_time)


def epoch_minutes_ago(n: int) -> int:
    """Returns the epoch in ms as it was exactly n-minutes ago."""
    date_time = datetime.datetime.now() - datetime.timedelta(minutes=n)
    return datetime_to_epoch_in_ms(date_time)


def epoch_hours_ago(n: int) -> int:
    """Returns the epoch in ms as it was exactly n-hours ago."""
    date_time = datetime.datetime.now() - datetime.timedelta(hours=n)
    return datetime_to_epoch_in_ms(date_time)


def epoch_days_ago(n: int) -> int:
    """Returns the epoch in ms as it was exactly n-days ago."""
    date_time = datetime.datetime.now() - datetime.timedelta(days=n)
    return datetime_to_epoch_in_ms(date_time)


def date_seconds_ago(n: int) -> int:
    """Returns the epoch in ms as it was exactly n-seconds ago."""
    date_time = datetime.datetime.now() - datetime.timedelta(minutes=n)
    return datetime_to_epoch_in_ms(date_time)


def date_minutes_ago(n: int) -> int:
    """Returns the epoch in ms as it was exactly n-minutes ago."""
    date_time = datetime.datetime.now() - datetime.timedelta(minutes=n)
    return datetime_to_epoch_in_ms(date_time)


def date_days_ago(n: int) -> int:
    """Returns the epoch in ms as it was exactly n-days ago."""
    date_time = datetime.datetime.now() - datetime.timedelta(days=n)
    return datetime_to_epoch_in_ms(date_time)


def epoch_ms_to_date_time(epoch_in_ms: int) -> datetime.datetime:
    """Take the number of milliseconds since epoch and return a datetime object"""
    epoch_time = epoch_in_ms / 1000.0
    return datetime.datetime.fromtimestamp(epoch_time)


def epoch_to_date_time(epoch_time: float) -> datetime.datetime:
    """Take the number of milliseconds since epoch and return a datetime object"""
    return datetime.datetime.fromtimestamp(epoch_time)


def convert_to_local(
    date: datetime.datetime, local_zone: Optional[str] = None
) -> datetime.datetime:
    """Convert UTC to a local date. If no local date is give, 'tz.tzlocal()' is used."""
    to_zone = local_zone if local_zone else tz.tzlocal()
    # date = date.replace(tzinfo=from_zone)
    converted = date.astimezone(to_zone)
    return converted


"""
def time_to_epoch_in_ms(
    year: int,
    month: int,
    day: int,
    hour: Optional[int],
    minute: Optional[int],
    second: Optional[int],
    microsecond: Optional[int],
    tzinfo: Optional[Union[None, datetime.tzinfo]],
) -> int:

    dttm = datetime.datetime(
        year, month, day, hour, minute, second, microsecond, tzinfo
    )
    print(dttm)
    pass
"""



import boto3
from dataclasses import dataclass
from ezaws.models.ssm import (
    GetParameterResponse,
    RegionParameters,
    CreateParameter,
    DeteleParameterResponse,
    CreateParameterResponse,
)


@dataclass
class ParameterStore:
    """
    Interface to the SSM Parameter store.
    """

    region: str

    def get_parameter(
        self, parameter: str, unencrypted: bool = True
    ) -> GetParameterResponse:
        """Retrieves target parameter, represented as a path to the parameter
        in SSM Parameter store."""
        ssm_client = boto3.client("ssm", self.region)

        client_response = ssm_client.get_parameter(
            Name=parameter, WithDecryption=unencrypted
        )

        return GetParameterResponse(**client_response)

    def describe_region_parameters(self) -> RegionParameters:
        """Returns information on a list of parameters in the region.
        Does not include the Value/secret itself."""
        ssm_client = boto3.client("ssm", self.region)

        client_response = ssm_client.describe_parameters()
        return RegionParameters(**client_response)

    def create_parameter(self, parameter: CreateParameter) -> CreateParameterResponse:
        """Create a parameter"""
        ssm_client = boto3.client("ssm", self.region)
        args = parameter.generate_parameter_args()

        client_response = ssm_client.put_parameter(**args)

        return CreateParameterResponse(**client_response)

    def delete_parameter(self, parameter_name: str) -> DeteleParameterResponse:
        """Delete a parameter from SSM parameter store."""
        ssm_client = boto3.client("ssm", self.region)
        arg = {"Name": parameter_name}
        client_response = ssm_client.delete_parameter(**arg)

        return DeteleParameterResponse(**client_response)



class EzawsBaseException(Exception):
    """General base exception except for exceptions that inherit from boto3."""

    pass


class CloudWatchException(EzawsBaseException):
    pass


class RDSException(EzawsBaseException):
    pass



import boto3
import botocore
from boto3.dynamodb.conditions import Key
from dataclasses import dataclass
from ezaws.models.dynamodb import (
    DeleteResponse,
    Table,
    GetItemResponse,
    ScanResult,
    PartiQLResult,
)
from typing import List, Any, Dict
import logging

logger = logging.getLogger(__name__)


@dataclass
class DynamoDB:
    """
    Interface to DynamoDB
    """

    region: str
    ddb_client: botocore.client = boto3.client("dynamodb")
    ddb_resource: boto3.resource = boto3.resource("dynamodb")

    def create_table(self, table: Table) -> bool:
        """Creates a DynamoDB table"""
        dynamodb = boto3.resource("dynamodb")

        key_schema_list = []
        for key in table.keys:
            key_schema_list.append(
                {"AttributeName": key.key_name, "KeyType": key.key_type}
            )
        attribute_list = []
        for attribute in table.attributes:
            attribute_list.append(
                {
                    "AttributeName": attribute.attribute_name,
                    "AttributeType": attribute.attribute_type,
                }
            )

        created_table = dynamodb.create_table(
            TableName=table.table_name,
            KeySchema=key_schema_list,
            AttributeDefinitions=attribute_list,
            ProvisionedThroughput={
                "ReadCapacityUnits": table.rcu,
                "WriteCapacityUnits": table.wcu,
            },  # 1 RCU and 1 WCU is free!
        )
        logger.info(
            f"Table {created_table.table_name} is in status {created_table.table_status}"
            "\nWaiting for the table to be fully provisioned."
        )

        created_table.meta.client.get_waiter("table_exists").wait(
            TableName=table.table_name
        )
        state = dynamodb.Table(table.table_name).table_status
        logger.info(f"Table created, status is now:{state}")

        return True

    def delete_table(self, table_name: str) -> DeleteResponse:
        """Deletes a DynamoDB table"""
        dynamodb = boto3.resource("dynamodb")

        table = dynamodb.Table(table_name)
        ret = table.delete()

        return DeleteResponse(**ret)

    def put_items(self, table_name: str, items: List[Dict[Any, Any]]) -> bool:
        """Add a list of items to target table."""

        for item in items:
            self.put_item(table_name=table_name, item=item)
        return True

    def put_item(self, table_name: str, item: Dict[Any, Any]) -> bool:
        """Add an item to target table."""
        table = self.ddb_resource.Table(table_name)
        table.put_item(Item=item)
        return True

    def scan(self, table_name: str) -> ScanResult:
        """Scan target table"""
        scan_result = self.ddb_client.scan(TableName=table_name)
        return ScanResult(**scan_result)

    def ez_scan(self, table_name: str) -> list[Any]:
        result = self.ddb_client.scan(table_name=table_name)
        deserializer = boto3.dynamodb.types.TypeDeserializer()

        def deserialize(d: dict) -> dict:
            return {k: deserializer.deserialize(v) for k, v in d.items()}

        data = [deserialize(d) for d in result.Items]
        return data

    def get_item(self, table_name: str, item: Dict[Any, Any]) -> GetItemResponse:
        """Get item from target table.

        Example:

        >>> ddb.get_item(
            table_name="Humans",
            item={"id": 1828, "name": "Arnetta"},
            )


        """
        table = self.ddb_resource.Table(table_name)
        get_item_result = table.get_item(Key=item)
        return GetItemResponse(**get_item_result)

    def list_tables(self) -> List[Any]:
        """Return a list of table objects (boto3.resources.factory.dynamodb.Table)
        for the region"""
        tables = list(self.ddb_resource.tables.all())
        return tables

    def list_table_names(self) -> List[str]:
        """Return a list of table names for the region"""
        tables = list(self.ddb_resource.tables.all())

        tables = [table.table_name for table in tables]

        return tables

    #    def query(self, table_name: str) -> QueryResult:
    #        dynamo_db = boto3.resource("dynamodb")
    #        table = dynamo_db.Table(table_name)
    #        response = table.query(KeyConditionExpression=Key("id").eq(10))
    #        result = QueryResult(**response)
    #        return result

    def partiql_query(self, partiql_query: str) -> PartiQLResult:
        """
        Execute a PartiQL query against Dynamo.

        For information on using PartiQL, check
         https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/ql-reference.html
        """
        logger.info(f"Executing partiql query {partiql_query}\n")
        result = self.ddb_client.execute_statement(Statement=partiql_query)
        return PartiQLResult(**result)



import boto3
import botocore

from dataclasses import dataclass
from botocore.exceptions import ClientError
from ezaws.models.rds import (
    DescribeDBResponse,
    DBInstanceNotFoundResponse,
    CreateDBResponse,
    DeleteDBResponse,
    Endpoint,
    DBInstance,
    DBEngine,
    StartRDSResponse,
    StopRDSResponse,
)
from typing import Union, Optional
from ezaws.exceptions import RDSException


@dataclass
class RDS:
    """
    Interface to the SSM Parameter store.
    """

    region: str
    master_user_password: str
    master_username: str
    db_name: str
    db_engine: DBEngine
    db_instance: DBInstance
    storage: int  # GiB
    arn: Optional[str] = None
    endpoint: Optional[Endpoint] = None
    rds_client: botocore.client = boto3.client("rds")

    def create_database(self) -> Union[DescribeDBResponse, CreateDBResponse]:
        """Create the database for this RDS instance"""
        resp = self.describe_db()
        if isinstance(resp, DescribeDBResponse):
            return resp
        elif isinstance(resp, DBInstanceNotFoundResponse):
            response = self.rds_client.create_db_instance(
                AllocatedStorage=self.storage,
                DBInstanceClass=self.db_instance,
                DBInstanceIdentifier=self.db_name,
                Engine=self.db_engine,
                MasterUserPassword=self.master_user_password,
                MasterUsername=self.master_username,
            )
            db_response = CreateDBResponse(**response)
            self.arn = db_response.DBInstance.DBInstanceArn
            return CreateDBResponse(**response)
        else:
            raise RDSException("Runtime error executing 'RDS.create_database'")

    def delete_database(self) -> Union[DeleteDBResponse, None]:
        """Delete this RDS instance"""
        resp = self.describe_db()
        if isinstance(resp, DescribeDBResponse):
            response = self.rds_client.delete_db_instance(
                DBInstanceIdentifier=self.db_name,
                SkipFinalSnapshot=True,
            )
            return DeleteDBResponse(**response)
        else:
            return None

    def describe_db(self) -> Union[DescribeDBResponse, DBInstanceNotFoundResponse]:
        """Have AWS describe the RDS instance, returns
        DBInstanceNotFoundResponse in case the instance does not exist."""

        try:
            response = self.rds_client.describe_db_instances(
                DBInstanceIdentifier=self.db_name,
                MaxRecords=100,
                Filters=[
                    {
                        "Name": "engine",
                        "Values": [
                            self.db_engine,
                        ],
                    }
                ],
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "DBInstanceNotFound":
                # this means the DB does not exist
                e.response
                return DBInstanceNotFoundResponse()
            raise e
        # pprint(response, width=2)
        return_value = DescribeDBResponse(**response)
        self.arn = return_value.DBInstances[0].DBInstanceArn
        if return_value.DBInstances[0].Endpoint:
            self.endpoint = return_value.DBInstances[0].Endpoint
        return return_value

    def start_db(self) -> StartRDSResponse:
        """Start the database instance"""
        response = self.rds_client.start_db_instance(DBInstanceIdentifier=self.db_name)

        return StartRDSResponse(**response)

    def stop_db(self, snap_shot_identifier: Optional[str] = None) -> StopRDSResponse:
        """Stop the database instance"""
        arg = {"DBInstanceIdentifier": self.db_name}
        if snap_shot_identifier:
            arg["DBSnapshotIdentifier"] = snap_shot_identifier
        response = self.rds_client.stop_db_instance(**arg)
        return StopRDSResponse(**response)
